#include <stdlib.h>
#include <stream.h>

/* CENSUS REPORT */
/* Include the SQL Communications Area. */

EXEC SQL
	INCLUDE SQLCA
/* Set up to catch SQL errors. Note: this will require that 
each routine containing executeable SQL commands has
an error handling section labeled 'err'.*/

EXEC SQL
	WHENEVER SQLERROR GO TO err;

/* Forward declarations */
static void print_error(), census_report();
int setup();

main()
{
	/* Set up estimated data; if successful generate report. */
	if (setup()) census_report();
	/* Undo estimated census data. */
	EXEC SQL
		ROLLBACK;
	exit (0);
	/* Output error information. */
err:		print_error();
	exit (1);
}
static void print_error()

/* Print out error message */
{
cout << "Data base error, SQLCODE = " << dec (SQLCODE, 5) << "\n";
}
int setup()
/* Put in estimated data for missing 1960 & 1970 census data. */
/* Return 1 if successful, 0 otherwise.  */
{
	/* Replace missing 1960 census data with estimates. */
	EXEC SQL
		UPDATE POPULATIONS
		SET CENSUS_1960 = 
			(0.3 * (CENSUS_1980 - CENSUS_1950)) + CENSUS_1950
			WHERE CENSUS_1960 IS NULL;
	/* Replace missing 1970 census data with estimates. */
	EXEC SQL
		UPDATE POPULATIONS
		SET CENSUS_1970 = 
			(0.65 * (CENSUS_1980 - CENSUS_1950)) + CENSUS_1950
			WHERE CENSUS_1970 IS NULL;
	return (1);
	/* Error return */
err:	print_error();
	return (0);
}
static void census_report()
{
	/* Declare variables to receive data retrieved by SQL
	commands. */
	BASED ON CITIES.POPULATION pop_50, pop_60, pop_70, pop_80;
	BASED ON STATES.STATE_NAME stname;
	/* Declare cursor for use in retrieving state/census data. */
  EXEC SQL
		DECLARE C CURSOR FOR
		SELECT STATE_NAME, CENSUS_1950, CENSUS_1960,
			CENSUS_1970, CENSUS_1980
		FROM STATES S, POPULATIONS P
		WHERE S.STATE = P.STATE
		ORDER S.STATE_NAME;
	/* Open cursor for use in retrieving state/census data. */
	EXEC SQL
		OPEN C;
	/* Print report header */
	cout << "     STATE                    "
	     << "                  POPULATION\n";
	cout << "                                1950        1960"
	     << "        1970        1980\n\n";
	/* Retrieve first set of state/census data. */
	EXEC SQL
		FETCH C INTO :stname, :pop_50, :pop_60, :pop_70, :pop_80;
	/* Until end of data stream is reached, print out current set
	   of data and then fetch next set. */
	while (SQLCODE == 0)
		{
                cout << form ("%-25s   %9d   %9d   %9d   %9d\n",
                        stname, pop_50, pop_60, pop_70, pop_80);
		/* Retrieve next set of state/census data. */
		EXEC SQL
			FETCH C INTO :stname, :pop_50, :pop_60,
				 :pop_70, :pop_80;
		}
	/* Close cursor. */
	EXEC SQL
		CLOSE C;
	return;
	/* Error return */
err:		print_error();
	return;
}
